/*
 * Copyright (C) 2008-2014 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define LOG_TAG "Sensors"

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <linux/input.h>
#include <math.h>
#include <poll.h>
#include <pthread.h>
#include <stdlib.h>

#include <utils/Atomic.h>
#include <utils/Log.h>

#include <hardware/sensors.h>

#include "sensors.h"
#include "CwMcuSensor.h"

/*****************************************************************************/

#define DELAY_OUT_TIME 0x7FFFFFFF

#define LIGHT_SENSOR_POLLTIME    2000000000

/*****************************************************************************/
static const struct sensor_t sSensorList[] = {
        { "Accelerometer Sensor",
                "HTC Group Ltd.",
                1, ID_A,
                SENSOR_TYPE_ACCELEROMETER, RANGE_A, CONVERT_A, 0.17f, 10000, 0, 0, 0, 0, 0, 0, {}},
        { "Magnetic field Sensor",
                "HTC Group Ltd.",
                1, ID_M,
                SENSOR_TYPE_MAGNETIC_FIELD, 200.0f, CONVERT_M, 5.0f, 16667, 0, 0, 0, 0, 0, 0, {}},
        { "Gyroscope Sensor",
                "HTC Group Ltd.",
                1, ID_GY,
                SENSOR_TYPE_GYROSCOPE, 40, CONVERT_GYRO, 6.1f, 1190, 0, 0, 0, 0, 0, 0, {}},
        {.name =       "CM32181 Light sensor",
         .vendor =     "Capella Microsystems",
         .version =    1,
         .handle =     ID_L,
         .type =       SENSOR_TYPE_LIGHT,
         .maxRange =   10240.0f,
         .resolution = 1.0f,
         .power =      0.15f,
         .minDelay =   0,
         .fifoReservedEventCount =  0,
         .fifoMaxEventCount =       0,
         .stringType = NULL,
         .requiredPermission = NULL,
         .maxDelay = 0,
         .flags = 0,
         .reserved = {}
        },

        { "Pressure Sensor",
                "HTC Group Ltd.",
                1, ID_PS,
                SENSOR_TYPE_PRESSURE, 2000, 1.0f, 0.0027f, 1190, 0, 0, 0, 0, 0, 0, {}},
        { "CWGD  Orientation Sensor",
                "HTC Group Ltd.",
                1, ID_O,
                SENSOR_TYPE_ORIENTATION, 360.0f, 0.1f, 11.27f, 10000, 0, 0, 0, 0, 0, 0, {}},
        { "Rotation Vector",
                "HTC Group Ltd.",
                1, ID_RV,
                SENSOR_TYPE_ROTATION_VECTOR, 1.0f, 1.0f / (1<<24), 11.27f, 1250, 0, 0, 0, 0, 0, 0, {}},
        { "Linear Acceleration",
                "HTC Group Ltd.",
                1, ID_LA,
                SENSOR_TYPE_LINEAR_ACCELERATION, RANGE_A, RESOLUTION_A, 11.27f, 10000, 0, 0, 0, 0, 0, 0, {}},
        { "Gravity",
                "HTC Group Ltd.",
                1, ID_G,
                SENSOR_TYPE_GRAVITY, GRAVITY_EARTH, (4.0f*9.81f)/256.0f, 11.27f, 10000, 0, 0, 0, 0, 0, 0, {}},
        { "Magnetic Uncalibrated",
                "hTC Corp.",
                1, ID_CW_MAGNETIC_UNCALIBRATED,
                SENSOR_TYPE_MAGNETIC_FIELD_UNCALIBRATED, 200.0f, 1.0f, 5.0f, 20000, 0, 0, 0, 0, 0, 0, {}},
        { "Gyroscope Uncalibrated",
                "hTC Corp.",
                1, ID_CW_GYROSCOPE_UNCALIBRATED,
                SENSOR_TYPE_GYROSCOPE_UNCALIBRATED, 200.0f, 1.0f, 6.1f, 20000, 0, 0, 0, 0, 0, 0, {}},
        { "Game Rotation Vector",
                "hTC Corp.",
                1, ID_CW_GAME_ROTATION_VECTOR,
                SENSOR_TYPE_GAME_ROTATION_VECTOR, 200.0f, 1.0f, 11.27f, 20000, 0, 0, 0, 0, 0, 0, {}},
        { "Geomagnetic Rotation Vector",
                "hTC Corp.",
                1, ID_CW_GEOMAGNETIC_ROTATION_VECTOR,
                SENSOR_TYPE_GEOMAGNETIC_ROTATION_VECTOR, 200.0f, 1.0f, 11.27f, 20000, 0, 0, 0, 0, 0, 0, {}},
        { "Significant Motion",
                "hTC Corp.",
                1, ID_CW_SIGNIFICANT_MOTION,
                SENSOR_TYPE_SIGNIFICANT_MOTION, 200.0f, 1.0f, 0.17f, 20000, 0, 0, 0, 0, 0, 0, {}},
        { "Step Detector",
                "hTC Corp.",
                1, ID_CW_STEP_DETECTOR,
                SENSOR_TYPE_STEP_DETECTOR, 200.0f, 1.0f, 0.17f, 0, 0, 0, 0, 0, 0, 0, {}},
        { "Step Counter",
                "hTC Corp.",
                1, ID_CW_STEP_COUNTER,
                SENSOR_TYPE_STEP_COUNTER, 200.0f, 1.0f, 0.17f, 0, 0, 0, 0, 0, 0, 0, {}},
        {.name =       "HTC Wake Gesture",
         .vendor =     "hTC Corp",
         .version =    1,
         .handle =     ID_WAKE_UP_GESTURE,
         .type =       SENSOR_TYPE_WAKE_GESTURE,
         .maxRange =   10240.0f,
         .resolution = 1.0f,
         .power =      0.17f,
         .minDelay =   0,
         .fifoReservedEventCount =  0,
         .fifoMaxEventCount =       0,
         .stringType = NULL,
         .requiredPermission = NULL,
         .maxDelay = 0,
         .flags = 0,
         .reserved = {}
        },
};

static int open_sensors(const struct hw_module_t* module, const char* id,
                        struct hw_device_t** device);

static int sensors__get_sensors_list(struct sensors_module_t*,
                                     struct sensor_t const** list)
{
    *list = sSensorList;
    return ARRAY_SIZE(sSensorList);
}

static struct hw_module_methods_t sensors_module_methods = {
    open: open_sensors
};

struct sensors_module_t HAL_MODULE_INFO_SYM = {
    common: {
            tag: HARDWARE_MODULE_TAG,
            version_major: 1,
            version_minor: 0,
            id: SENSORS_HARDWARE_MODULE_ID,
            name: "Sensor module",
            author: "Electronic Company",
            methods: &sensors_module_methods,
            dso: NULL,
            reserved: { },
    },
    get_sensors_list: sensors__get_sensors_list,
};

struct sensors_poll_context_t {
    sensors_poll_device_1_t device; // must be first

        sensors_poll_context_t();
        ~sensors_poll_context_t();
    int activate(int handle, int enabled);
    int setDelay(int handle, int64_t ns);
    int pollEvents(sensors_event_t* data, int count);
    int batch(int handle, int flags, int64_t period_ns, int64_t timeout);
    int flush(int handle);

private:
    enum {
        cwmcu            = 0,
        numSensorDrivers,
        numFds,
    };

    static const size_t wake = numFds - 1;
    static const char WAKE_MESSAGE = 'W';
    struct pollfd mPollFds[numFds];
    int mWritePipeFd;
    SensorBase* mSensors[numSensorDrivers];

int handleToDriver(int handle) const {
        switch (handle) {
                case ID_A:
                case ID_M:
                case ID_GY:
                case ID_L:
                case ID_PS:
                case ID_O:
                case ID_RV:
                case ID_LA:
                case ID_G:
                case ID_CW_MAGNETIC_UNCALIBRATED:
                case ID_CW_GYROSCOPE_UNCALIBRATED:
                case ID_CW_GAME_ROTATION_VECTOR:
                case ID_CW_GEOMAGNETIC_ROTATION_VECTOR:
                case ID_CW_SIGNIFICANT_MOTION:
                case ID_CW_STEP_DETECTOR:
                case ID_CW_STEP_COUNTER:
                case ID_WAKE_UP_GESTURE:
                        return cwmcu;
        }
        return -EINVAL;
    }
};

/*****************************************************************************/

sensors_poll_context_t::sensors_poll_context_t()
{
    mSensors[cwmcu] = new CwMcuSensor();
    mPollFds[cwmcu].fd = mSensors[cwmcu]->getFd();
    mPollFds[cwmcu].events = POLLIN;
    mPollFds[cwmcu].revents = 0;

    int wakeFds[2];
    int result = pipe(wakeFds);
    ALOGE_IF(result<0, "error creating wake pipe (%s)", strerror(errno));
    fcntl(wakeFds[0], F_SETFL, O_NONBLOCK);
    fcntl(wakeFds[1], F_SETFL, O_NONBLOCK);
    mWritePipeFd = wakeFds[1];

    mPollFds[wake].fd = wakeFds[0];
    mPollFds[wake].events = POLLIN;
    mPollFds[wake].revents = 0;
}

sensors_poll_context_t::~sensors_poll_context_t() {
    for (int i=0 ; i<numSensorDrivers ; i++) {
        delete mSensors[i];
    }
    close(mPollFds[wake].fd);
    close(mWritePipeFd);
}

int sensors_poll_context_t::activate(int handle, int enabled) {
    int index = handleToDriver(handle);
    if (index < 0) return index;
    int err =  mSensors[index]->setEnable(handle, enabled);
    if (enabled && !err) {
        const char wakeMessage(WAKE_MESSAGE);
        int result = write(mWritePipeFd, &wakeMessage, 1);
        ALOGE_IF(result<0, "error sending wake message (%s)", strerror(errno));
    }
    return err;
}

int sensors_poll_context_t::setDelay(int handle, int64_t ns) {

    int index = handleToDriver(handle);
    if (index < 0) return index;
    return mSensors[index]->setDelay(handle, ns);
}

int sensors_poll_context_t::pollEvents(sensors_event_t* data, int count)
{
    int nbEvents = 0;
    int n = 0;
    do {
        // see if we have some leftover from the last poll()
        for (int i=0 ; count && i<numSensorDrivers ; i++) {
            SensorBase* const sensor(mSensors[i]);
            if ((mPollFds[i].revents & POLLIN) || (sensor->hasPendingEvents())) {
                int nb = sensor->readEvents(data, count);
                if (nb < count) {
                    // no more data for this sensor
                    mPollFds[i].revents = 0;
                }
                count -= nb;
                nbEvents += nb;
                data += nb;
            }
        }

        if (count) {
            // we still have some room, so try to see if we can get
            // some events immediately or just wait if we don't have
            // anything to return
            do {
                TEMP_FAILURE_RETRY(n = poll(mPollFds, numFds, nbEvents ? 0 : -1));
            } while (n < 0 && errno == EINTR);
            if (n<0) {
                ALOGE("poll() failed (%s)", strerror(errno));
                return -errno;
            }
            if (mPollFds[wake].revents & POLLIN) {
                char msg(WAKE_MESSAGE);
                int result = read(mPollFds[wake].fd, &msg, 1);
                ALOGE_IF(result<0, "error reading from wake pipe (%s)", strerror(errno));
                ALOGE_IF(msg != WAKE_MESSAGE, "unknown message on wake queue (0x%02x)", int(msg));
                mPollFds[wake].revents = 0;
            }
        }
        // if we have events and space, go read them
    } while (n && count);
    return nbEvents;
}

int sensors_poll_context_t::batch(int handle, int flags, int64_t period_ns, int64_t timeout)
{
    int index = handleToDriver(handle);

    if (index < 0)
        return index;

    int err = mSensors[index]->batch(handle, flags, period_ns, timeout);

    return err;
}

int sensors_poll_context_t::flush(int handle)
{
    int index = handleToDriver(handle);

    if (index < 0)
        return index;

    int err = mSensors[index]->flush(handle);

    return err;
}


/*****************************************************************************/

static int poll__close(struct hw_device_t *dev)
{
    sensors_poll_context_t *ctx = reinterpret_cast<sensors_poll_context_t *>(dev);
    if (ctx) {
        delete ctx;
    }
    return 0;
}

static int poll__activate(struct sensors_poll_device_t *dev,
        int handle, int enabled) {
    sensors_poll_context_t *ctx = reinterpret_cast<sensors_poll_context_t *>(dev);
    return ctx->activate(handle, enabled);
}

static int poll__setDelay(struct sensors_poll_device_t *dev,
        int handle, int64_t ns) {
    sensors_poll_context_t *ctx = reinterpret_cast<sensors_poll_context_t *>(dev);
    return ctx->setDelay(handle, ns);
}

static int poll__poll(struct sensors_poll_device_t *dev,
        sensors_event_t* data, int count) {
    sensors_poll_context_t *ctx = reinterpret_cast<sensors_poll_context_t *>(dev);
    return ctx->pollEvents(data, count);
}

static int poll__batch(struct sensors_poll_device_1 *dev,
                      int handle, int flags, int64_t period_ns, int64_t timeout)
{
    sensors_poll_context_t *ctx = (sensors_poll_context_t *)dev;
    return ctx->batch(handle, flags, period_ns, timeout);
}

static int poll__flush(struct sensors_poll_device_1 *dev,
                      int handle)
{
    sensors_poll_context_t *ctx = (sensors_poll_context_t *)dev;
    return ctx->flush(handle);
}
/*****************************************************************************/

// Open a new instance of a sensor device using name
static int open_sensors(const struct hw_module_t* module, const char*,
                        struct hw_device_t** device)
{
    sensors_poll_context_t *dev = new sensors_poll_context_t();

    memset(&dev->device, 0, sizeof(sensors_poll_device_1_t));

    dev->device.common.tag = HARDWARE_DEVICE_TAG;
    dev->device.common.version  = SENSORS_DEVICE_API_VERSION_1_1;
    dev->device.common.module   = const_cast<hw_module_t*>(module);
    dev->device.common.close    = poll__close;
    dev->device.activate        = poll__activate;
    dev->device.setDelay        = poll__setDelay;
    dev->device.poll            = poll__poll;

    // Batch processing
    dev->device.batch           = poll__batch;
    dev->device.flush           = poll__flush;

    *device = &dev->device.common;

    return 0;
}

